# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import signal
import subprocess
from distutils import spawn

import fixtures

import testtools


class TestCli(testtools.TestCase):

    @testtools.skipUnless(spawn.find_executable("memcached"),
                          "memcached not found")
    def test_cli(self):
        self.assertEqual(0, os.system(
            "pifpaf run memcached --port 11216 echo >/dev/null 2>&1"))

    @staticmethod
    def _read_stdout_and_kill(stdout):
        env = {}
        for line in stdout.split(b'\n'):
            k, _, v = line.partition(b"=")
            env[k] = v
        os.kill(int(env[b"export PIFPAF_PID"].strip(b"\" \n;")),
                signal.SIGTERM)
        return env

    @testtools.skipUnless(spawn.find_executable("memcached"),
                          "memcached not found")
    def test_eval(self):
        c = subprocess.Popen(["pifpaf", "run", "memcached", "--port", "11219"],
                             stdout=subprocess.PIPE)
        (stdout, stderr) = c.communicate()
        self.assertEqual(0, c.wait())
        env = self._read_stdout_and_kill(stdout)

        self.assertEqual(b"\"memcached://localhost:11219\";",
                         env[b"export PIFPAF_URL"])
        self.assertEqual(b"\"memcached://localhost:11219\";",
                         env[b"export PIFPAF_MEMCACHED_URL"])

    @testtools.skipUnless(spawn.find_executable("memcached"),
                          "memcached not found")
    def test_exit_code(self):
        c = subprocess.Popen(["pifpaf", "run", "memcached", "--port", "11234",
                              "--", "bash", "-c", "exit 31"],
                             stdout=subprocess.PIPE)
        (stdout, stderr) = c.communicate()
        self.assertEqual(31, c.wait())

    @testtools.skipUnless(spawn.find_executable("memcached"),
                          "memcached not found")
    def test_env_prefix(self):
        c = subprocess.Popen(["pifpaf", "run",
                              "--env-prefix", "FOOBAR",
                              "memcached", "--port", "11215"],
                             bufsize=0,
                             stdout=subprocess.PIPE)
        (stdout, stderr) = c.communicate()
        self.assertEqual(0, c.wait())
        env = self._read_stdout_and_kill(stdout)

        self.assertEqual(b"\"memcached://localhost:11215\";",
                         env[b"export FOOBAR_URL"])
        self.assertEqual(b"\"memcached://localhost:11215\";",
                         env[b"export FOOBAR_MEMCACHED_URL"])
        self.assertEqual(env[b"export PIFPAF_PID"],
                         env[b"export FOOBAR_PID"])

    @testtools.skipUnless(spawn.find_executable("memcached"),
                          "memcached not found")
    def test_env_prefix_old_format(self):
        # Old format
        c = subprocess.Popen(["pifpaf",
                              "--env-prefix", "FOOBAR",
                              "run", "memcached", "--port", "11215"],
                             bufsize=0,
                             stdout=subprocess.PIPE)
        (stdout, stderr) = c.communicate()
        self.assertEqual(0, c.wait())
        env = self._read_stdout_and_kill(stdout)

        self.assertEqual(b"\"memcached://localhost:11215\";",
                         env[b"export FOOBAR_URL"])
        self.assertEqual(b"\"memcached://localhost:11215\";",
                         env[b"export FOOBAR_MEMCACHED_URL"])
        self.assertEqual(env[b"export PIFPAF_PID"],
                         env[b"export FOOBAR_PID"])

    @testtools.skipUnless(spawn.find_executable("memcached"),
                          "memcached not found")
    def test_global_urls_varibale(self):
        c = subprocess.Popen(["pifpaf", "run",
                              "--env-prefix", "FOOBAR",
                              "memcached", "--port", "11217",
                              "--",
                              "pifpaf",
                              "run", "memcached", "--port", "11218"],
                             bufsize=0,
                             stdout=subprocess.PIPE)
        (stdout, stderr) = c.communicate()
        self.assertEqual(0, c.wait())
        env = self._read_stdout_and_kill(stdout)

        self.assertEqual(b"\"memcached://localhost:11218\";",
                         env[b"export PIFPAF_URL"])
        self.assertEqual(b"\"memcached://localhost:11218\";",
                         env[b"export PIFPAF_MEMCACHED_URL"])
        self.assertEqual(
            b"\"memcached://localhost:11217;memcached://localhost:11218\";",
            env[b"export PIFPAF_URLS"])

    @testtools.skipUnless(spawn.find_executable("memcached"),
                          "memcached not found")
    def test_global_urls_varibale_old_format(self):
        c = subprocess.Popen(["pifpaf",
                              "--env-prefix", "FOOBAR",
                              "run", "memcached", "--port", "11217",
                              "--",
                              "pifpaf",
                              "run", "memcached", "--port", "11218"],
                             bufsize=0,
                             stdout=subprocess.PIPE)
        (stdout, stderr) = c.communicate()
        self.assertEqual(0, c.wait())
        env = self._read_stdout_and_kill(stdout)

        self.assertEqual(b"\"memcached://localhost:11218\";",
                         env[b"export PIFPAF_URL"])
        self.assertEqual(b"\"memcached://localhost:11218\";",
                         env[b"export PIFPAF_MEMCACHED_URL"])
        self.assertEqual(
            b"\"memcached://localhost:11217;memcached://localhost:11218\";",
            env[b"export PIFPAF_URLS"])

    def test_non_existing_command(self):
        # Keep PATH to just the one set by tox to run pifpaf
        self.useFixture(fixtures.EnvironmentVariable(
            "PATH", os.getenv("PATH").split(":")[0]))
        c = subprocess.Popen(["pifpaf", "run", "memcached"],
                             bufsize=0,
                             stderr=subprocess.PIPE,
                             stdout=subprocess.PIPE)
        (stdout, stderr) = c.communicate()
        self.assertEqual(1, c.wait())
        self.assertIn(
            b"ERROR [pifpaf] Unable to run command "
            b"`memcached -p 11212': [Errno 2] No such file or directory",
            stderr)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from pifpaf import drivers


class QdrouterdDriver(drivers.Driver):
    DEFAULT_PORT = 5672
    DEFAULT_ARTEMIS_PORT = 5673
    DEFAULT_USERNAME = "pifpaf"
    DEFAULT_PASSWORD = "secrete"
    DEFAULT_DOMAIN = "localhost"

    def __init__(self, port=DEFAULT_PORT,
                 artemis_port=DEFAULT_ARTEMIS_PORT,
                 username=DEFAULT_USERNAME,
                 password=DEFAULT_PASSWORD,
                 domain=DEFAULT_DOMAIN,
                 mesh=False,
                 direct_notify=False,
                 **kwargs):
        """Create a new Qdrouterd instance."""
        super(QdrouterdDriver, self).__init__(templatedir="qdrouterd",
                                              **kwargs)
        self.port = port
        self.artemis_port = artemis_port
        self.username = username
        self.password = password
        self.domain = domain
        self.mesh = mesh
        self.direct_notify = direct_notify

    @classmethod
    def get_options(cls):
        return [
            {"param_decls": ["--port"],
             "type": int,
             "default": cls.DEFAULT_PORT,
             "help": "port to use for Qdrouterd"},
            {"param_decls": ["--artemis_port"],
             "type": int,
             "default": cls.DEFAULT_ARTEMIS_PORT,
             "help": "port to use for broker link"},
            {"param_decls": ["--mesh"],
             "is_flag": True,
             "help": "TODO: Create a 3 HA node mesh"},
            {"param_decls": ["--direct_notify"],
             "is_flag": True,
             "help": "direct message notify and do not attach to broker"},
            {"param_decls": ["--username"],
             "default": cls.DEFAULT_USERNAME,
             "help": "sasl username"},
            {"param_decls": ["--password"],
             "default": cls.DEFAULT_PASSWORD,
             "help": "sasl password"},
            {"param_decls": ["--domain"],
             "default": cls.DEFAULT_DOMAIN,
             "help": "sasl domain"},
        ]

    def saslpasswd2(self, username, password, sasl_db):
        self._exec(["saslpasswd2", "-c", "-p", "-f",
                    sasl_db, username], stdin=password.encode())

    def _setUp(self):
        super(QdrouterdDriver, self)._setUp()

        # setup log, etc used by qdrouterd
        logdir = os.path.join(self.tempdir, "log")
        os.makedirs(logdir)
        etcdir = os.path.join(self.tempdir, "etc")
        os.makedirs(etcdir)
        sasldir = os.path.join(etcdir, "sasl2")
        os.makedirs(sasldir)
        logfile = os.path.join(logdir, "qdrouterd.log")

        template_env = {
            "TMP_DIR": self.tempdir,
            "PORT": self.port,
            "ARTEMIS_PORT": self.artemis_port,
            "SASL_DIR": sasldir,
            "LOG_FILE": logfile,
            "DIRECT_NOTIFY": self.direct_notify,
        }

        qdr_cfg = os.path.join(etcdir, "qdrouterd.conf")
        self.template("qdrouterd.conf",
                      template_env,
                      qdr_cfg)

        sasl_cfg = os.path.join(sasldir, "sasl_qdrouterd.conf")
        self.template("sasl_qdrouterd.conf",
                      template_env,
                      sasl_cfg)

        sasl_db = os.path.join(sasldir, "qdrouterd.sasldb")
        self.saslpasswd2(self.username, self.password, sasl_db)

        c, _ = self._exec(["qdrouterd", "-c", qdr_cfg],
                          wait_for_port=self.port)

        self.putenv("QDROUTERD_PORT", str(self.port))
        self.putenv("QDROUTERD_URL", "amqp://localhost:%s" % self.port)
        self.putenv("URL", "amqp://%s:%s@localhost:%s" % (
            self.username, self.password, self.port))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from pifpaf import drivers


class KafkaDriver(drivers.Driver):
    DEFAULT_KAFKA_PORT = 9092
    DEFAULT_ZOOKEEPER_PORT = 2181
    DEFAULT_PATH = ["/opt/kafka/bin",
                    "/usr/local/opt/kafka/bin"]

    def __init__(self, port=DEFAULT_KAFKA_PORT,
                 zookeeper_port=DEFAULT_ZOOKEEPER_PORT,
                 **kwargs):
        """Create a new Kafka instance."""
        super(KafkaDriver, self).__init__(**kwargs)
        self.port = port
        self.zookeeper_port = zookeeper_port

    def _setUp(self):
        super(KafkaDriver, self)._setUp()

        suffix = ".sh"
        if self.find_executable("zookeeper-server-start", self.DEFAULT_PATH):
            suffix = ""

        # This is use explicitly byu kafka AND implicitly by zookeeper
        logdir = os.path.join(self.tempdir, "log")
        os.makedirs(logdir)

        zookeeper_conf = os.path.join(self.tempdir, "zookeeper.properties")
        kafka_conf = os.path.join(self.tempdir, "kafka.properties")

        with open(zookeeper_conf, "w") as f:
            f.write("""
dataDir=%s
clientPort=%s
maxClientCnxns=0
""" % (self.tempdir, self.zookeeper_port))

        with open(kafka_conf, "w") as f:
            f.write("""
broker.id=0
advertised.listeners=PLAINTEXT://localhost:%d
num.network.threads=3
num.io.threads=8
socket.send.buffer.bytes=102400
socket.receive.buffer.bytes=102400
socket.request.max.bytes=104857600
log.dirs=%s
num.partitions=2
num.recovery.threads.per.data.dir=1
log.retention.hours=168
log.segment.bytes=1073741824
log.retention.check.interval.ms=300000
zookeeper.connect=localhost:%d
zookeeper.connection.timeout.ms=6000
offsets.topic.replication.factor=1
transaction.state.log.replication.factor=1
transaction.state.log.min.isr=1
group.initial.rebalance.delay.ms=0
""" % (self.port, logdir, self.zookeeper_port))

        # NOTE(sileht): The wait_for_line is the best we can do
        # but we error can occur after the last line we see when it works...
        env = {"LOG_DIR": logdir}
        self._exec(['zookeeper-server-start%s' % suffix, zookeeper_conf],
                   wait_for_line='binding to port .*:%s' % self.zookeeper_port,
                   path=self.DEFAULT_PATH, env=env,
                   forbidden_line_after_start=(2, "Unexpected exception"))
        # We ignore failure because stop script kill all zookeeper pids
        # (even the system one)
        self.addCleanup(self._exec, ['zookeeper-server-stop%s' % suffix],
                        path=self.DEFAULT_PATH, env=env, ignore_failure=True)

        self._exec(['kafka-server-start%s' % suffix, kafka_conf],
                   wait_for_line='[Kafka Server id=0].*started',
                   path=self.DEFAULT_PATH, env=env,
                   forbidden_line_after_start=(2,
                                               "kafka.common.KafkaException"))
        self.addCleanup(self._exec, ['kafka-server-stop%s' % suffix],
                        path=self.DEFAULT_PATH, env=env, ignore_failure=True)

        self.putenv("KAFKA_PORT", str(self.port))
        self.putenv("KAFKA_PROTOCOL", "PLAINTEXT")
        self.putenv("KAFKA_URL", "PLAINTEXT://localhost:%s" % self.port)
        self.putenv("URL", "kafka://localhost:%s" % self.port)
